#include <ch32v00x.h>

static const uint32_t frequencies[] = {
	// 1 = Do4, 2 = Ré4, 3 = Mi4,
	523, 587, 659,
	// 4 = Fa4, 5 = Sol4, 6 = La4,
	698, 783, 880,
	// 7 = Si4, 8 = Do5, 9 = Ré5,
	987, 1046, 1174,
	// * = Mi5, 0 = Fa5, # = Sol5,
	1318, 1396, 1567,
};

int main() {
	// On initialise le GPIO
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_GPIOC, ENABLE);
	GPIO_InitTypeDef gpioInit = { 0 };
	// PC0..3 correspondent aux rangées (entrées), de haut en bas.
	gpioInit.GPIO_Pin = GPIO_Pin_0 | GPIO_Pin_1 | GPIO_Pin_2 | GPIO_Pin_3;
	gpioInit.GPIO_Mode = GPIO_Mode_IPD;
	GPIO_Init(GPIOC, &gpioInit);
	// PC5..7 correspondent aux colonnes (sorties), de gauche à droite.
	gpioInit.GPIO_Pin = GPIO_Pin_5 | GPIO_Pin_6 | GPIO_Pin_7;
	gpioInit.GPIO_Mode = GPIO_Mode_Out_PP;
	gpioInit.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_Init(GPIOC, &gpioInit);
	// PC4 est la sortie PWM (T1CH4).
	gpioInit.GPIO_Pin = GPIO_Pin_4;
	gpioInit.GPIO_Mode = GPIO_Mode_AF_PP;
	GPIO_Init(GPIOC, &gpioInit);

	RCC_APB2PeriphClockCmd(RCC_APB2Periph_TIM1, ENABLE);

	TIM_TimeBaseInitTypeDef timeBaseInit = { 0 };
	// La configuration du prescaler sera faite après le démarrage du compteur.
	//timeBaseInit.TIM_Prescaler = 0;
	timeBaseInit.TIM_Period = 99;
	timeBaseInit.TIM_ClockDivision = TIM_CKD_DIV1;
	timeBaseInit.TIM_CounterMode = TIM_CounterMode_Up;
	TIM_TimeBaseInit(TIM1, &timeBaseInit);

	TIM_OCInitTypeDef ocInit = { 0 };
	ocInit.TIM_OCMode = TIM_OCMode_PWM1;
	ocInit.TIM_OCPolarity = TIM_OCPolarity_Low;
	ocInit.TIM_OutputState = TIM_OutputState_Enable;
	ocInit.TIM_Pulse = 50;
	TIM_OC4Init(TIM1, &ocInit);

	TIM_CtrlPWMOutputs(TIM1, ENABLE);
	// Tout est en place, mais on ne démarrera le compteur que
	// lorsqu'on aura une note à jouer.

	int note = -1;
	int currentColumn = 0;

	while (1) {
		// Le principe est d'envoyer un niveau logique haut alternativement
		// sur chaque sortie. Si on retrouve ce niveau sur une entrée, c'est
		// que la touche à l'intersection des deux est appuyée.
		switch (currentColumn) {
		case 0:
			GPIO_Write(GPIOC, 0x20);
			break;

		case 1:
			GPIO_Write(GPIOC, 0x40);
			break;

		case 2:
			GPIO_Write(GPIOC, 0x80);
			break;
		}

		// La lecture du port renvoie l'état de toutes les lignes,
		// donc sorties comprises.
		int key = GPIO_ReadInputData(GPIOC);

		// On regarde si on trouve un niveau logique haut sur une des entrées.
		if (key & 0x0f) {
			// On recherche la position du "1" dans les entrées pour trouver
			// la rangée.
			int row = 0;
			for (int v = key; !(v & 1); v >>= 1, row++);
			// On recherche la position du "1" dans les sorties pour trouver
			// la colonne, en ignorant PC4 que nous n'utilisons pas.
			int column = 0;
			for (int v = key; !(v & 0x20); v >>= 1, column++);

			int newNote = 3 * row + column;

			// Inutile de toucher au timer si rien n'a changé.
			if (newNote != note) {
				note = newNote;
				// On joue jusqu'à nouvel ordre.
				TIM_Cmd(TIM1, ENABLE);
				// Configurer le prescaler pour rendre la note souhaitée.
				TIM1->PSC = ((48000000ul / 100ul) / frequencies[note]) - 1;
			}
		} else if (note >= 0) {
			// On a relâché la touche, mettre fin à la note en cours.
			note = -1;
			TIM_Cmd(TIM1, DISABLE);
		}

		// On passe à la colonne suivante.
		currentColumn++;

		if (currentColumn > 2) {
			currentColumn = 0;
		}
	}
}
